import unittest
from unittest.mock import MagicMock, call, patch
from parameterized import parameterized

import numpy as np
from numpy.testing import assert_array_equal
from pathlib import Path
from txrm2tiff.txrm.main import open_txrm
from txrm2tiff.txrm import annot_mixin
from PIL import Image, ImageDraw

from txrm2tiff.xradia_properties.enums import AnnotationTypes, XrmDataTypes


test_data_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/")
test_file = test_data_path / "annotation_test" / "Xray_mosaic_F3C.xrm"

visit_path = test_data_path / "data" / "2019" / "cm98765-1"
raw_path = visit_path / "raw"
xm10_path = raw_path / "XMv10"
xm13_path = raw_path / "XMv13"

test_files = [
    (test_file,),
    (xm13_path / "Xray_mosaic_v13.xrm",),
    (xm13_path / "Xray_mosaic_v13_interrupt.xrm",),
    (xm13_path / "Xray_mosaic_7x7_v13.xrm",),
    (xm13_path / "Xray_single_v13.xrm",),
    (xm13_path / "tomo_v13_full.txrm",),
    (xm13_path / "tomo_v13_full_noref.txrm",),
    (xm13_path / "tomo_v13_interrupt.txrm",),
    (xm13_path / "VLM_mosaic_v13.xrm",),
    (xm13_path / "VLM_mosaic_v13_interrupt.xrm",),
    (xm13_path / "VLM_grid_mosaic_large_v13.xrm",),
]


class TestAnnotator(unittest.TestCase):
    @parameterized.expand(test_files)
    @unittest.skipUnless(visit_path.exists(), "dls paths cannot be accessed")
    def test_with_real_image(self, test_file):
        with open_txrm(test_file) as txrm:
            annotations = txrm.extract_annotations()
        self.assertTrue(np.any(annotations))

    def test_square(self):
        fill = 125
        im_size = 5
        imshape = (1, im_size, im_size)

        x0, y0 = 1, 1
        x1, y1 = 4, 4

        ann = annot_mixin.AnnotatorMixin()
        ann.read_stream = MagicMock()
        ann.output_shape = imshape

        image = Image.new("RGB", imshape[:0:-1], (0, 0, 0))
        draw = ImageDraw.Draw(image, mode="RGB")

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                ann.read_stream.side_effect = [[x0], [y0], [x1], [y1]]
                stream_stem = ""
                ann._plot_rect(draw, stream_stem)

        expected_output = np.zeros((im_size, im_size, 3), dtype=np.uint8)
        y0, y1 = im_size - 1 - y0, im_size - 1 - y1  # Invert y axis
        expected_output[y0, x0 : x1 + 1, 1] = fill
        expected_output[y1 : y0 + 1, x0, 1] = fill
        expected_output[y1, x0 : x1 + 1, 1] = fill
        expected_output[y1 : y0 + 1, x1, 1] = fill

        assert_array_equal(np.asarray(image), expected_output)

    def test_thickness_modifier(self):
        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = (6, 500, 1500)
        self.assertEqual(ann.thickness_modifier, 2.0)

    def test_line(self):
        fill = 125
        x0, x1 = 1, 4
        y = 4
        im_size = 5
        imshape = (1, im_size, im_size)
        im = Image.new("RGBA", imshape[:0:-1], (0, 0, 0, 0))
        draw = ImageDraw.Draw(im, mode="RGBA")

        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = imshape
        ann.read_stream = MagicMock()

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                ann.read_stream.side_effect = [[x0], [x1], [y], [y]]
                ann._plot_line(draw, "")

        output = np.asarray(im.convert("RGB"))

        expected_output = np.zeros((im_size, im_size, 3), dtype=np.uint8)
        y = im_size - 1 - y  # Invert y axis
        expected_output[y, x0 : x1 + 1, 1] = fill

        assert_array_equal(output, expected_output)

    def test_line_and_apply(self):
        fill = (0, 125, 0)
        x0, x1 = 1, 4
        y = 4
        im_size = 5
        im_arr = np.zeros((1, im_size, im_size))
        annotations = Image.new("RGBA", im_arr.shape[:0:-1], (0, 0, 0, 0))
        draw = ImageDraw.Draw(annotations, mode="RGBA")

        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = im_arr.shape
        ann.read_stream = MagicMock()

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (*fill, 255)
                patched_thickness.return_value = 1
                ann.read_stream.side_effect = [[x0], [x1], [y], [y]]
                ann._plot_line(draw, "")

        output = ann.apply_annotations(im_arr, annotations)

        expected_output = np.zeros((1, im_size, im_size, 3), dtype=np.uint8)
        y = im_size - 1 - y  # Invert y axis
        expected_output[0][y, x0 : x1 + 1, :] = fill

        assert_array_equal(output, expected_output)

    def test_ellipse(self):
        fill = 125
        xs = 0, 18
        ys = 2, 16
        x_mid = int(round(np.mean(xs)))
        y_mid = int(round(np.mean(ys)))
        im_size = 20
        imshape = (1, im_size, im_size)
        im = Image.new("RGBA", imshape[:0:-1], (0, 0, 0, 0))
        draw = ImageDraw.Draw(im, mode="RGBA")

        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = imshape
        ann.read_stream = MagicMock()

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                ann.read_stream.side_effect = [[xs[0]], [ys[0]], [xs[1]], [ys[1]]]
                ann._plot_ellipse(draw, "")

        output = np.asarray(im)

        self.assertTrue((output[:, :, 0] == 0).any())  # Red
        self.assertTrue((output[:, :, 2] == 0).any())  # Blue
        for x, y in ann._flip_y(
            (x_mid, ys[0]), (xs[0], y_mid), (x_mid, ys[1]), (xs[1], y_mid)
        ):
            self.assertEqual(
                output[y, x, 1], fill, msg="Failed for y, x of %i, %i" % (y, x)
            )

    def test_polyline(self):
        fill = 125
        im_size = 18  # Must be an even number
        xs = tuple(range(im_size))
        ys = [0, im_size - 1] * (im_size // 2)
        total_points = len(xs)

        imshape = (1, im_size, im_size)
        im = Image.new("RGBA", imshape[:0:-1], (0, 0, 0, 0))
        draw = ImageDraw.Draw(im, mode="RGBA")

        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = imshape
        ann.read_stream = MagicMock()

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                ann.read_stream.side_effect = [[total_points], xs, ys]
                ann._plot_polyline(draw, "")

        output = np.asarray(im)

        self.assertTrue((output[:, :, 0] == 0).any())  # Red
        self.assertTrue((output[:, :, 2] == 0).any())  # Blue
        for x, y in ann._flip_y(*zip(xs, ys)):
            self.assertEqual(output[y, x, 1], fill)

    def test_scale_bar(self):
        im_size = 5
        imshape = (1, im_size, im_size)
        im = Image.new("RGBA", imshape[:0:-1], (0, 0, 0, 0))
        draw = ImageDraw.Draw(im, mode="RGBA")

        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = imshape
        ann.image_info = {"PixelSize": [17.0]}
        self.assertTrue(ann._add_scale_bar(draw), msg="Failed to add scale bar")

        output = np.asarray(im)

        self.assertTrue((output[:, :, 0] == 0).any(), msg="You have red on you")  # Red
        self.assertTrue(
            (output[:, :, 1] == 255).any(), msg="You don't have green on you"
        )  # Green
        self.assertTrue(
            (output[:, :, 2] == 0).any(), msg="You have blue on you"
        )  # Blue

    def test_flip_y(self):
        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = (20, 20)
        res = ann._flip_y((0, 0), (19, 19))
        self.assertEqual(res, ((0, 19), (19, 0)))

    def test_create_image_and_draw(self):
        z, x, y = 2, 6, 5
        channels = 4  # RGBA
        ann = annot_mixin.AnnotatorMixin()
        ann.output_shape = (z, y, x)
        ann.has_stream = MagicMock(return_value=False)
        im, draw = ann._create_image_and_draw()
        self.assertEqual(np.asarray(im).shape, (y, x, channels))

    def test_annotate(self):
        class TestAnnotator(annot_mixin.AnnotatorMixin):
            @property
            def thickness_modifier(self) -> float:
                return 1

        fill = (0, 125, 0)
        x0, x1 = 1, 4
        y = 4
        im = np.zeros((5, 6, 7), dtype=np.uint8)  # Z, Y, X
        im[0, 0, 0] = 255  # Ensure range stays the same as it is cast to 'L' (uint8)
        im = np.flip(im, axis=1)
        num_annotations = 1
        ann = TestAnnotator()
        ann.get_output = MagicMock(return_value=im)
        ann.output_shape = im.shape
        ann.has_stream = MagicMock(return_value=True)
        ann.read_stream = MagicMock(
            side_effect=[  # Draw a single line with the listed points
                [num_annotations],
                [AnnotationTypes.ANN_LINE.value],
                [x0],
                [x1],
                [y],
                [y],
            ]
        )
        ann._add_scale_bar = MagicMock(return_value=True)

        with patch.object(ann, "_get_colour") as patched_colour, patch.object(
            ann, "_get_thickness"
        ) as patched_thickness:
            patched_colour.return_value = (*fill, 255)
            patched_thickness.return_value = 1

            ann.annotate()

        ann.read_stream.assert_has_calls(
            [
                call("Annot/TotalAnn", strict=True),
                call("Annot/Ann0/AnnType", XrmDataTypes.XRM_INT, strict=True),
                call("Annot/Ann0/X1", XrmDataTypes.XRM_FLOAT, strict=True),
                call("Annot/Ann0/X2", XrmDataTypes.XRM_FLOAT, strict=True),
                call("Annot/Ann0/Y1", XrmDataTypes.XRM_FLOAT, strict=True),
                call("Annot/Ann0/Y2", XrmDataTypes.XRM_FLOAT, strict=True),
            ]
        )

        # Extra axis of length 3 for RGB
        expected_output = np.stack([im] * 3, axis=3)
        y = expected_output.shape[1] - 1 - y  # Invert y axis
        expected_output[:, y, x0 : x1 + 1, :] = fill
        self.assertEqual(
            ann.annotated_image.shape, expected_output.shape, msg="Shapes don't match"
        )
        unequal_pos = np.where(ann.annotated_image != expected_output)
        assert_array_equal(
            ann.annotated_image,
            expected_output,
            err_msg=f"{[(x, y, z, c) for z, y, x, c in zip(*unequal_pos)]}:\n{ann.annotated_image[unequal_pos]}\n\n{expected_output[unequal_pos]}",
        )



import typing
from collections import defaultdict
from os import PathLike
import olefile
from io import IOBase

from .abstract import AbstractTxrm
from .v3 import Txrm3
from .v5 import Txrm5
from .. import txrm_functions
from ..utils import file_handler


def __invalid_txrm_file(file: PathLike, *args) -> None:
    raise IOError(f"Invalid txrm file '{file}'")


txrm_classes = defaultdict(
    lambda: __invalid_txrm_file,
    {
        3: Txrm3,
        5: Txrm5,
    },
)


def open_txrm(
    file: typing.Union[str, PathLike, IOBase, bytes],
    load_images: bool = True,
    load_reference: bool = True,
    strict: bool = False,
) -> typing.Optional[AbstractTxrm]:
    TxrmClass = get_txrm_class(file)
    return TxrmClass(file, load_images, load_reference, strict)


def get_txrm_class(
    file: typing.Union[str, PathLike, IOBase, bytes],
) -> typing.Optional[AbstractTxrm]:
    if isinstance(file, (str, PathLike)) and (
        not file_handler.file_can_be_opened(file)
        or not file_handler.ole_file_works(file)
    ):
        return __invalid_txrm_file(file)
    try:
        ole = olefile.OleFileIO(file)
        m_version = int(txrm_functions.get_file_version(ole))
    finally:
        if not isinstance(file, IOBase):
            # Don't close the IO as it can't be reopened
            ole.close()
    return txrm_classes[m_version]



import unittest
from unittest.mock import patch, MagicMock

import os
import sys
from pathlib import Path
from subprocess import Popen, PIPE, STDOUT

from txrm2tiff import __version__
from txrm2tiff import __main__


class TestCommandlineEntryPoint(unittest.TestCase):
    # Before these running tests, the module must be installed in development mode using:
    # `pip install --editable . --user` (from the project folder)
    # it can be uninstalled at the end using:
    # `pip uninstall txrm2tiff`
    # IF TESTING ON WINDOWS: make sure the correct python installation is set as the default in PATH

    @classmethod
    def setUpClass(cls):
        # Set maximum difference string length to None (infinite)
        cls.maxDiff = None
        # Add local intall to path for testing with development install
        if os.name == "posix":
            os.environ["PATH"] += os.pathsep + os.path.join(
                os.path.expanduser("~"), ".local", "bin"
            )

    @patch("argparse.ArgumentParser.print_help")
    def test_argparse_function(self, mocked_printhelp):
        sys.argv = ["txrm2tiff"]
        __main__.main()

        mocked_printhelp.assert_called()

    @patch("argparse.ArgumentParser.print_help")
    def test_argparse_function_setup_subpaser(self, mocked_printhelp):
        sys.argv = ["txrm2tiff", "setup"]
        __main__.main()

        mocked_printhelp.assert_called_once_with()

    def test_argparse_function_setup_windows_shortcut(self):
        with patch(
            "txrm2tiff.utils.shortcut_creation.create_Windows_shortcut", MagicMock()
        ) as mocked_shortcut_creation:
            sys.argv = ["txrm2tiff", "setup", "-w"]
            __main__.main()

            mocked_shortcut_creation.assert_called_once_with()

    def test_argparse_function_with_args(self):
        with patch("txrm2tiff.convert_and_save", MagicMock()) as mock_convert_and_save:
            mock_convert_and_save.return_value = None
            input_arg = "input_path"
            annotate_arg = False
            flip = False
            ref_arg = None
            data_type_arg = None
            sys.argv = ["txrm2tiff", "--input", input_arg]
            __main__.main()

            mock_convert_and_save.assert_called_once_with(
                input_path=input_arg,
                custom_reference=ref_arg,
                output_path=None,
                annotate=annotate_arg,
                flip=flip,
                data_type=data_type_arg,
                apply_shifts=False,
                ignore_reference=False,
                logging_level="info",
            )

    def test_argparse_function_with_all_args(self):
        with patch("txrm2tiff.convert_and_save", MagicMock()) as mock_convert_and_save:
            mock_convert_and_save.return_value = None
            input_arg = "input_path"
            annotate_arg = True
            flip_arg = True
            ref_arg = "ref_path"
            data_type_arg = "uint16"
            sys.argv = [
                "txrm2tiff",
                "--input",
                input_arg,
                "--reference",
                ref_arg,
                "--datatype",
                data_type_arg,
                "--annotate",
                "--flip",
                "--apply-shifts",
                "--ignore-ref",
            ]
            __main__.main()

            mock_convert_and_save.assert_called_once_with(
                input_path=input_arg,
                custom_reference=ref_arg,
                output_path=None,
                annotate=annotate_arg,
                flip=flip_arg,
                data_type=data_type_arg,
                apply_shifts=True,
                ignore_reference=True,
                logging_level="info",
            )

    def test_script_method(self):
        args = [Path("path_to/input.txrm")]
        run_args = ["txrm2tiff", "-i", str(args[0])]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(
            f"No such file or directory: {args[0]}",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )

    def test_script_method_version_number(self):
        run_args = ["txrm2tiff", "--version"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        # Get the last line (ignore the input)
        stdout = [l.strip() for l in stdout.strip("\n").split("\n")][-1]
        self.assertEqual(
            f"txrm2tiff {__version__}", stdout, msg=f"Actual stdout: {stdout}"
        )

    def test_script_method_setup_subparser(self):
        run_args = ["txrm2tiff", "setup"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.replace("\r\n", " ").replace("\n", " ")
        self.assertIn(
            "usage: txrm2tiff setup [-w] [-h]", stdout, msg=f"Actual stdout: {stdout}"
        )

    def test_script_method_inspect_subparser(self):
        run_args = ["txrm2tiff", "inspect"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = " ".join(
            [text.strip() for text in stdout.replace("\r\n", " ").split("\n")]
        )
        self.assertIn(
            "txrm2tiff inspect [-i INPUT_PATH] [-e] [-l] [-s STREAMS [STREAMS ...]] [-h]",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )

    def test_module_without_arguments_returns_help(self):
        run_args = [sys.executable, "-m", "txrm2tiff"]
        path = os.environ["PATH"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(
            "Converter of txrm/xrm files to OME tif/tiff files",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )

    def test_module_function_with_args(self):
        input_path = "input_file_path"
        custom_reference = "ref_path"
        output_path = None
        annotate = True
        flip = True
        data_type = None
        apply_shifts = False
        ignore_reference = False
        logging_level = 1

        run_args = [
            sys.executable,
            "-m",
            "txrm2tiff",
            "--input",
            input_path,
            "--reference",
            custom_reference,
            "--set-logging",
            str(logging_level),
            "--annotate",
            "--flip",
        ]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.replace("\r\n", " ").replace("\n", " ")
        self.assertIn(
            f"Running with arguments: input_path={input_path}, custom_reference={custom_reference}, output_path={output_path}, annotate={annotate}, flip={flip}, data_type={data_type}, apply_shifts={apply_shifts}, ignore_reference={ignore_reference}, logging_level={logging_level}",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )
        self.assertIn(
            f"No such file or directory: {input_path}",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )

    def test_module_method(self):
        input_path = Path("path_to/input.txrm")
        custom_reference = None
        output_path = None
        annotate = False
        flip = False
        data_type = None
        apply_shifts = False
        ignore_reference = False
        logging_level = 1
        args = [input_path]
        run_args = [
            sys.executable,
            "-m",
            "txrm2tiff",
            "-i",
            str(args[0]),
            "--set-logging",
            str(logging_level),
        ]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.replace("\r\n", " ").replace("\n", " ")
        self.assertIn(
            f"Running with arguments: input_path={input_path}, custom_reference={custom_reference}, output_path={output_path}, annotate={annotate}, flip={flip}, data_type={data_type}, apply_shifts={apply_shifts}, ignore_reference={ignore_reference}, logging_level={logging_level}",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )
        self.assertIn(
            f"No such file or directory: {input_path}",
            stdout,
            msg=f"Actual stdout: {stdout}",
        )

    def test_module_method_version_number(self):
        run_args = [sys.executable, "-m", "txrm2tiff", "--version"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        # Get the last line (ignore the input)
        stdout = [l.strip() for l in stdout.strip("\n").split("\n")][-1]
        self.assertEqual(
            f"txrm2tiff {__version__}", stdout, msg=f"Actual stdout: {stdout}"
        )

    def test_module_method_setup_subparser_without_arguments_returns_help(self):
        run_args = [sys.executable, "-m", "txrm2tiff", "setup"]
        with Popen(run_args, stdout=PIPE, stderr=STDOUT, universal_newlines=True) as p:
            stdout, _ = p.communicate()
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(
            f"txrm2tiff setup [-w] [-h]", stdout, msg=f"Actual stdout: {stdout}"
        )



import typing
import logging
import numpy as np

from .abstract import AbstractTxrm
from .ref_mixin import ReferenceMixin
from .annot_mixin import AnnotatorMixin
from .save_mixin import SaveMixin
from .shifts_mixin import ShiftsMixin
from .txrm_property import txrm_property
from ..txrm_functions.images import fallback_image_interpreter
from ..utils.image_processing import stitch_images


class Txrm5(ShiftsMixin, SaveMixin, ReferenceMixin, AnnotatorMixin, AbstractTxrm):
    @txrm_property(fallback=None)
    def is_mosaic(self) -> bool:
        return bool(np.sum(self.mosaic_dims))

    @txrm_property(fallback=[])
    def image_dims(self) -> typing.List[int]:
        return [
            self.image_info["ImageWidth"][0],
            self.image_info["ImageHeight"][0],
        ]

    @txrm_property(fallback=[])
    def reference_dims(self):
        return [
            self.reference_info["ImageWidth"][0],
            self.reference_info["ImageHeight"][0],
        ]

    @txrm_property(fallback=None)
    def reference_exposure(self) -> typing.Optional[float]:
        if "ExpTimes" in self.reference_info:
            return self.reference_info["ExpTimes"][0]
        elif "ExpTime" in self.reference_info:
            return self.reference_info["ExpTime"][0]

    def extract_reference_image(self) -> np.ndarray:
        ref_data = self.extract_reference_data()

        if isinstance(ref_data, bytes):  # If unable to extract dtype
            ref_data = fallback_image_interpreter(
                ref_data, np.prod(self.reference_dims), self.strict
            )

        return ref_data.reshape(self.reference_dims[::-1])

    def get_output(
        self,
        load: bool = False,
        shifts: bool = False,
        flip: bool = False,
        clear_images: bool = True,
    ) -> typing.Optional[np.ndarray]:
        """
        Returns output image as ndarray with axes [idx, y, x]. If a reference has been applied, the referenced image will be returned.

        load: load the image(s) if they are not already loaded. Does not apply any reference.
        flip: flip the Y-axis of the output image(s) (how they are displayed in DX)
        clear_images: clear images and reference from the Txrm instance after returning.
        """
        images = self.get_images(load)
        if images is None:
            logging.warning("No image has been loaded, so no output can be returned.")
            return None
        images = images.copy()
        if shifts and self.has_shifts:
            images = self.apply_shifts_to_images(images)
        if self.is_mosaic:
            images = stitch_images(images, self.mosaic_dims)
        if clear_images:
            self.clear_images()
            self.clear_reference()
        if not flip:
            # The default state is flipped with respect to how it's displayed in XRM Data Explorer
            return np.flip(images, axis=1)
        return images



import logging
from typing import List
import numpy as np
from numpy.typing import DTypeLike, NDArray

from .functions import convert_to_int


def dynamic_despeckle_and_average_series(
    images: np.ndarray, average: bool = True
) -> np.ndarray:
    """
    Despeckle and average a series of images (requires min 3 images)
    This uses the same or similar logic to Xradia/Zeiss' API.

    Args:
        images (list): List of 2D numpy arrays
        average (bool, optional): Defaults to True.

    Returns:
        numpy array: single image (2D array) containing despeckle and averaged data
    """
    nimages, height, width = images.shape
    if nimages < 3:
        logging.error("Despeckle averaging requires a minimum of 3 images")
        return None
    # Takes a list of XradiaData.FloatArray as input
    # outputs a 2D ndarray
    block = images.astype(np.float32).reshape((nimages, height * width))
    # Sort pixels stack-wise
    vals = np.sort(block, axis=0)
    # Set number of sorted frames to split off
    maligned_idxs = [nimages - 1]
    if nimages >= 9:
        maligned_idxs.extend([nimages - 2, nimages - 3, 0])
        good_vals = vals[1:-3]
    else:
        good_vals = vals[:-1]
    maligned = vals[maligned_idxs, :]
    # Calculate mean and sigma of 'good' pixels
    mean = np.mean(good_vals, axis=0, keepdims=True)
    sigma = np.std(good_vals, axis=0, ddof=1, keepdims=True)
    # Create boolean mask for cutting out 'bad' pixels
    bad_pixels = np.fabs(maligned - mean) > 2.8 * sigma
    # Set bad pixels to nan and apply to vals
    maligned[bad_pixels] = np.nan
    vals[maligned_idxs] = maligned
    # Get stack-wise mean (ignoring invalid nans)
    output_image = np.nanmean(vals, axis=0, dtype=np.float32)
    if not average:
        output_image *= nimages
    return output_image.reshape(height, width)


def normalise_to_datatype(
    array: np.ndarray, datatype: DTypeLike, clip: bool = False
) -> np.ndarray:
    logging.debug("Re-scaling array to %s", datatype)
    array = array.astype(np.float64)  # Convert to float for rescaling

    # Clip outliers before normalisation to avoid any speckles
    # that can lead to the normalised image to be dark:
    if clip:
        num_std = 3.0
        logging.info("Clipping outliers (>%g std from mean)", num_std)
        minimum = 0
        maximum = np.mean(array) + num_std * np.std(array)
    else:
        minimum = array.min()
        maximum = array.max()

    return rescale_image(
        array,
        np.iinfo(datatype).min,
        np.iinfo(datatype).max,
        previous_minimum=minimum,
        previous_maximum=maximum,
    )


def rescale_image(
    array, minimum, maximum, previous_minimum=None, previous_maximum=None
):
    if previous_minimum is None:
        previous_minimum = np.min(array)
    if previous_maximum is None:
        previous_maximum = np.max(array)

    return np.interp(
        np.clip(array, a_min=previous_minimum, a_max=previous_maximum),
        (previous_minimum, previous_maximum),
        (minimum, maximum),
    )


def cast_to_dtype(image: np.ndarray, data_type: DTypeLike) -> np.ndarray:
    try:
        dtype = np.dtype(data_type)
        if dtype is image.dtype:
            logging.debug("Image is already %s", dtype)
        else:
            if np.issubdtype(dtype, np.integer):
                dtype_info = np.iinfo(dtype)
            elif np.issubdtype(dtype, np.floating):
                dtype_info = np.finfo(dtype)
            else:
                raise TypeError(f"Cannot cast to invalid data type {data_type}")
            # Round min/max to avoid this warning when the issue is just going to be rounded away.
            img_min, img_max = image.min(), image.max()
            dtype_min, dtype_max = dtype_info.min, dtype_info.max
            if dtype_min > img_min:
                logging.warning(
                    "Image min %f below %s minimum of %i, values below this will be cut off",
                    img_min,
                    dtype,
                    dtype_min,
                )
            else:
                dtype_min = None
            if dtype_max < img_max:
                logging.warning(
                    "Image max %f above %s maximum of %i, values above this will be cut off",
                    img_max,
                    dtype,
                    dtype_max,
                )
            else:
                dtype_max = None

            if dtype_min is not None or dtype_max is not None:
                np.clip(image, dtype_min, dtype_max, out=image)

            if np.issubdtype(dtype, np.integer) and np.issubdtype(
                image.dtype, np.floating
            ):
                image = np.around(image, decimals=0)

            image = image.astype(dtype, copy=False)
            logging.info("Image has been cast to %s", data_type)
    except TypeError:
        logging.warning(
            "Invalid data type given: %s aka %s. Image will remain as %s.",
            data_type,
            dtype,
            image.dtype,
        )
    except Exception:
        logging.error(
            "An error occurred casting image from %s to %s", image.dtype, data_type
        )
        raise
    return image


def tile_image_data_to_mosaic(
    refdata: np.ndarray,
    image_rows: int,
    image_columns: int,
    mosaic_rows: int,
    mosaic_columns: int,
) -> np.ndarray:
    ref_num_rows = convert_to_int(image_rows / mosaic_rows)
    ref_num_columns = convert_to_int(image_columns / mosaic_columns)
    refdata.shape = (ref_num_rows, ref_num_columns)
    return np.tile(refdata, (mosaic_rows, mosaic_columns))


def stitch_images(images: np.ndarray, mosaic_dims: List[int]) -> np.ndarray:
    """
    Stitches images into a mosaic stored as a 2D array.

    Mosaic dims should be X, Y.
    """
    # Convert to Y X for numpy:
    slow_np_axis = 0
    fast_np_axis = 1

    num_images, y, x = images.shape
    expected_num_images = mosaic_dims[0] * mosaic_dims[1]

    if expected_num_images > num_images:
        images.resize((expected_num_images, y, x), refcheck=False)

    fast_stacked_list = []
    for i in range(mosaic_dims[slow_np_axis]):
        idx_start = mosaic_dims[fast_np_axis] * i
        idx_end = mosaic_dims[fast_np_axis] * (i + 1)
        logging.debug(
            "Stitching mosaic images %i to %i along axis %s",
            idx_start,
            (idx_end - 1),
            ("x", "y")[fast_np_axis],
        )
        fast_stacked_list.append(
            np.concatenate(images[idx_start:idx_end], axis=fast_np_axis)
        )
    # Must be output in a 3D numpy array (with axis 0 as z)
    final_image = np.concatenate(fast_stacked_list, axis=slow_np_axis)[np.newaxis, :, :]
    logging.info("Mosaic stitched into image with shape: %s", final_image.shape)
    return final_image



from datetime import datetime
import unittest
from unittest.mock import MagicMock, patch
from numpy.testing import assert_array_equal

import numpy as np

from txrm2tiff.txrm.abstract import AbstractTxrm, txrm_property
from txrm2tiff.xradia_properties.enums import XrmDataTypes


@patch.multiple(
    AbstractTxrm, __abstractmethods__=set(), open=MagicMock
)  # Allows abstract method to be initialised and file to not be opened
class AbstractTxrmTest(unittest.TestCase):
    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"ImagesTaken": [500]}
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.shape", new=(6, 9))
    def test_extract_images(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        images = [
            np.arange(a, a + np.prod(txrm.shape))
            for a in np.arange(txrm.image_info["ImagesTaken"][0])
        ]
        for im in images:
            im.shape = txrm.shape
        with patch.object(txrm, "_extract_single_image") as mocked_extract_single_image:
            mocked_extract_single_image.side_effect = images
            output = txrm.extract_images()

        expected = np.asarray(images)

        assert_array_equal(output, expected)

    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"ImagesTaken": [500]}
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.shape", new=(6, 9))
    def test_extract_images_no_images(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        with patch.object(txrm, "_extract_single_image") as mocked_extract_single_image:
            mocked_extract_single_image.side_effect = TypeError("Test")
            output = txrm.extract_images()

        self.assertTrue(output.size == 0)

    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"ImagesTaken": [500]}
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.shape", new=(6, 9))
    def test_extract_images_no_images_strict(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        txrm.strict = True
        with patch.object(txrm, "_extract_single_image") as mocked_extract_single_image:
            mocked_extract_single_image.side_effect = TypeError("Test")
            with self.assertRaises(TypeError):
                txrm.extract_images()

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.shape", new=(6, 9))
    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_dtype",
        XrmDataTypes.XRM_UNSIGNED_INT,
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.has_stream")
    @patch("txrm2tiff.txrm.abstract.txrm_functions.get_stream_from_bytes")
    def test_extract_single_image_known_dtype(
        self, mocked_frombytes, mocked_has_stream
    ):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        image = np.arange(np.prod(txrm.shape))
        stream_data = image.tobytes()
        stream = MagicMock()
        stream.getvalue.return_value = stream_data
        txrm.ole = MagicMock()
        txrm.ole.openstream.return_value = stream

        mocked_frombytes.return_value = image
        output = txrm._extract_single_image(201)
        mocked_frombytes.assert_called_once_with(
            stream_data, dtype=txrm.image_dtype.value
        )
        mocked_has_stream.assert_called_once_with("ImageData3/Image201")
        image.shape = txrm.shape
        assert_array_equal(output, image)

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.shape", new=(6, 9))
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.has_stream")
    @patch("txrm2tiff.txrm.abstract.txrm_functions.fallback_image_interpreter")
    def test_extract_single_image_unknown_dtype(
        self,
        mocked_interpreter,
        mocked_has_stream,
    ):
        mocked_has_stream.return_value = True
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        imsize = np.prod(txrm.shape)
        image = np.arange(imsize)
        stream_data = image.tobytes()
        stream = MagicMock()
        stream.getvalue.return_value = stream_data
        txrm.ole = MagicMock()
        txrm.ole.openstream.return_value = stream

        mocked_interpreter.return_value = image
        output = txrm._extract_single_image(201)
        mocked_interpreter.assert_called_once_with(stream_data, imsize, txrm.strict)
        mocked_has_stream.assert_called_once_with("ImageData3/Image201")
        image.shape = txrm.shape
        assert_array_equal(output, image)

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.shape", new=(6, 9))
    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_dtype",
        XrmDataTypes.XRM_UNSIGNED_INT,
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.has_stream")
    @patch("txrm2tiff.txrm.abstract.txrm_functions.get_stream_from_bytes")
    @patch("txrm2tiff.txrm.abstract.txrm_functions.fallback_image_interpreter")
    def test_extract_single_image_fails(
        self,
        mocked_interpreter,
        mocked_frombytes,
        mocked_has_stream,
    ):
        mocked_has_stream.return_value = True
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        imsize = np.prod(txrm.shape)
        image = np.arange(imsize)
        stream_data = image.tobytes()
        stream = MagicMock()
        stream.getvalue.return_value = stream_data
        txrm.ole = MagicMock()
        txrm.ole.openstream.return_value = stream

        empty_img = np.asarray([])

        mocked_frombytes.side_effect = TypeError("Test")
        mocked_interpreter.return_value = empty_img
        output = txrm._extract_single_image(201)
        mocked_interpreter.assert_called_once_with(stream_data, imsize, txrm.strict)
        mocked_has_stream.assert_called_once_with("ImageData3/Image201")
        image.shape = txrm.shape
        assert_array_equal(output, empty_img)

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.get_images")
    def test_get_single_image_is_loaded(self, mocked_get_images):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        img = "test_img"
        images = [img]
        mocked_get_images.return_value = images
        output = txrm.get_single_image(1)
        mocked_get_images.assert_called_once_with(load=False)
        self.assertEqual(output, img)

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.get_images")
    def test_get_single_image_is_loaded(self, mocked_get_images):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        img = "test_img"
        images = [img]
        mocked_get_images.return_value = images
        output = txrm.get_single_image(1)
        mocked_get_images.assert_called_once_with(load=False)
        self.assertEqual(output, img)

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.get_images")
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm._extract_single_image")
    def test_get_single_image_is_not_loaded(
        self, mocked_extract_single, mocked_get_images
    ):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        img = "test_img"
        images = [img]
        mocked_extract_single.return_value = img
        mocked_get_images.return_value = None
        output = txrm.get_single_image(1)
        mocked_get_images.assert_called_once_with(load=False)
        self.assertEqual(output, img)

    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.reference_dtype",
        XrmDataTypes.XRM_UNSIGNED_INT,
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.has_reference", new=True)
    @patch("txrm2tiff.txrm.abstract.txrm_functions.get_stream_from_bytes")
    def test_extract_reference_data(self, mocked_frombytes):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        stream = MagicMock()
        stream.getvalue.return_value = "ref_data"
        txrm.ole = MagicMock()
        txrm.ole.openstream.return_value = stream

        mocked_frombytes.return_value = "output"
        output = txrm.extract_reference_data()
        mocked_frombytes.assert_called_once_with("ref_data", dtype=txrm.reference_dtype.value)
        self.assertEqual(output, "output")

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.has_reference", new=True)
    @patch("txrm2tiff.txrm.abstract.txrm_functions.get_stream_from_bytes")
    def test_extract_reference_data_unknown_dtype_returns_bytes(self, mocked_frombytes):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        stream = MagicMock()
        stream.getvalue.return_value = "ref_data"  # This would be bytes in a real case
        txrm.ole = MagicMock()
        txrm.ole.openstream.return_value = stream
        output = txrm.extract_reference_data()
        mocked_frombytes.assert_not_called()
        self.assertEqual(output, "ref_data")

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"ImagesTaken": [97]})
    def test_get_central_image(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        with patch.object(txrm, "get_single_image") as mocked_get_single_image:
            txrm.get_central_image()
            mocked_get_single_image.assert_called_once_with(49)

    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_info",
        new={"ImagesTaken": [500]},
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.is_mosaic", new=False)
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.image_dims", new=[6, 7])
    def test_output_shape_non_mosaic(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        self.assertEqual(txrm.output_shape, [500, 7, 6])

    @patch(
        "txrm2tiff.txrm.abstract.AbstractTxrm.image_info",
        new={"ImagesTaken": [6], "MosiacColumns": [2], "MosiacRows": [3]},
    )
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.image_dims", new=[6, 7])
    def test_output_shape_mosaic(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        self.assertEqual(txrm.output_shape, [1, 7 * 3, 6 * 2])

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"Date": ["ksfs$oio12/30/2021 23:55:59!j#f"]})
    def test_datetimes(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        self.assertEqual(txrm.datetimes[0], datetime(2021, 12, 30, 23, 55, 59))

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"Date": ["ksfs$oio12/30/21 23:55:59!j#f"]})
    def test_datetimes_without_century(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        self.assertEqual(txrm.datetimes[0], datetime(2021, 12, 30, 23, 55, 59))

    @patch("txrm2tiff.txrm.abstract.AbstractTxrm.image_info", new={"Date": ["ksfs$oio12/30/2021 23:55:59.83!j#f"]})
    def test_datetimes_with_milliseconds(self):
        txrm = AbstractTxrm("test/path", load_images=False, load_reference=False)
        self.assertEqual(txrm.datetimes[0], datetime(2021, 12, 30, 23, 55, 59, 830000))



from os import PathLike
from pathlib import Path
import logging
from typing import Optional, Union
from numpy.typing import DTypeLike

from .txrm import open_txrm
from .txrm.abstract import AbstractTxrm
from .utils.logging import create_logger


def convert_and_save(
    input_path: Union[str, PathLike],
    output_path: Optional[Union[str, PathLike]] = None,
    custom_reference: Optional[Union[str, PathLike]] = None,
    annotate: bool = False,
    flip: bool = False,
    data_type: Optional[str] = None,
    apply_shifts: bool = False,
    ignore_reference: bool = False,
    logging_level: Union[str, int] = "info",
) -> None:
    """
    Converts xrm or txrm file then saves as it as .ome.tiff file.

    Args:
        input_path (str): Path, can be a file or directory (a directory will batch convert all valid contents).
        custom_reference (str, optional): File path, not valid for batch processing. Defaults to None.
        output_path (str, optional): Output path, for batch processing output must be a directory otherwise outputs will be saved in the same directory as each input file found. Defaults to None.
        annotate (bool, optional): Save additional annotated image (if annotations are available). Defaults to False.
        apply_shifts (bool, optional): Apply saved shifts (if available)
        ignore_reference (bool, optional): Ignore any internal reference. Defaults to False.
        logging_level (str, optional): Defaults to "info".
    """
    create_logger(str(logging_level).lower())
    logging.debug(
        "Running with arguments: "
        "input_path=%s, custom_reference=%s, output_path=%s, annotate=%s, flip=%s, data_type=%s, apply_shifts=%s, ignore_reference=%s, logging_level=%s",
        input_path,
        custom_reference,
        output_path,
        annotate,
        flip,
        data_type,
        apply_shifts,
        ignore_reference,
        logging_level,
    )
    input_path = Path(input_path)
    if input_path.exists():
        if output_path is not None:
            output_path = Path(output_path)
        if custom_reference is not None:
            logging.warning("Custom references are invalid for batch conversions")
        if input_path.is_dir():
            logging.info("Converting files in directory '%s'", input_path)
            _batch_convert_files(
                input_path,
                output_path,
                annotate,
                flip,
                data_type,
                apply_shifts,
                ignore_reference,
            )
        else:
            if custom_reference is not None:
                custom_reference = Path(custom_reference)
                if not custom_reference.is_file():
                    logging.warning(
                        "The specified custom reference file '%s' does not exist.",
                        custom_reference,
                    )
                    custom_reference = None
            logging.info("Converting file '%s'", input_path)
            _convert_and_save(
                input_path,
                output_path,
                custom_reference,
                annotate,
                flip,
                data_type,
                apply_shifts,
                ignore_reference,
            )
    else:
        logging.error("No such file or directory: %s", input_path)
        raise IOError(f"No such file or directory: {input_path}")


def _batch_convert_files(
    input_directory: Path,
    output: Optional[Path] = None,
    annotate: bool = True,
    flip: bool = False,
    data_type: Optional[DTypeLike] = None,
    shifts: bool = False,
    ignore_reference: bool = False,
) -> None:
    filepath_list = _find_files(input_directory)
    logging.info(
        f"Batch converting files: {', '.join([filepath.name for filepath in filepath_list])}"
    )
    if output is not None:
        logging.info(f"Output directory set to: {output}")
        output_path_base = Path(output)
        # Replace input_filepath with output, maintaining sub directory structure, and find output file suffix
        output_path_list = [
            _set_output_suffix(output_path_base / filepath.relative_to(input_directory))
            for filepath in filepath_list
        ]
        for output_path in output_path_list:
            # Make any directories that do not exist:
            output_path.parent.mkdir(parents=True, exist_ok=True)
    else:
        # Defines output now rather than in _convert_and_save so that there's a list to iterate
        output_path_list = [_set_output_suffix(filepath) for filepath in filepath_list]

    for filepath, output_path in zip(filepath_list, output_path_list):
        _convert_and_save(
            filepath,
            output_path,
            None,
            annotate,
            flip,
            data_type,
            shifts,
            ignore_reference,
        )


def _find_files(directory: Path):
    return list(directory.rglob("*xrm"))


def _convert_and_save(
    input_path: Path,
    output_path: Optional[Path] = None,
    custom_reference: Optional[Union[str, PathLike]] = None,
    annotate: bool = True,
    flip: bool = False,
    data_type: Optional[DTypeLike] = None,
    shifts: bool = False,
    ignore_reference: bool = False,
) -> None:
    with open_txrm(input_path) as txrm:
        _convert_file(txrm, custom_reference, ignore_reference, annotate)
        output_path = _decide_output_path(txrm.path, output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)  # Make output directory
        txrm.save_images(output_path, data_type, flip=flip, shifts=shifts, mkdir=True)


def _convert_file(
    txrm: AbstractTxrm,
    custom_reference: Optional[Union[str, PathLike]] = None,
    ignore_reference: bool = False,
    annotate: bool = True,
) -> None:
    logging.info(f"Converting {txrm.name}")
    # custom_reference overrules ignore_reference
    if custom_reference is not None or not ignore_reference:
        txrm.apply_reference(
            custom_reference
        )  # Called with None applies internal reference
    if annotate:
        txrm.annotate()


def _decide_output_path(input_path: Path, output_path: Optional[Path]) -> Path:
    # If no output is supplied
    if output_path is None:
        output_path = input_path
    else:
        if output_path.suffix == "" or output_path.is_dir():
            output_path = output_path / input_path.name
        else:
            # If a valid filepath was given, just return that (avoids double .ome)
            return output_path
    return _set_output_suffix(output_path, current_suffix=input_path.suffix)


def _set_output_suffix(filepath: Path, current_suffix: Optional[str] = None) -> Path:
    if current_suffix is None:
        current_suffix = filepath.suffix
    if current_suffix.lower() in (".txrm", ".xrm"):
        return filepath.with_suffix(".ome.tiff")
    else:
        logging.error("Invalid file extension: %s", current_suffix)
        raise NameError(f"Invalid file extension: {current_suffix}")



import logging
import tifffile as tf
import numpy as np
from numpy.typing import DTypeLike
from os import access, R_OK, PathLike
from pathlib import Path
from typing import Optional, Union, List
from olefile import OleFileIO, isOleFile
from oxdls import OMEXML

from .metadata import dtype_dict
from .image_processing import cast_to_dtype
from ..txrm_functions import read_stream
from ..xradia_properties import XrmDataTypes as XDT
from ..info import __version__


def file_can_be_opened(path: Union[str, PathLike]) -> bool:
    if access(str(path), R_OK):
        return True
    logging.error("File %s cannot be opened", path)
    return False


def ole_file_works(path: Union[str, PathLike]) -> bool:
    path = Path(path)
    if path.is_file() and ((path.suffix == ".txrm") or (path.suffix == ".xrm")):
        if isOleFile(str(path)):
            with OleFileIO(str(path)) as ole_file:
                number_frames_taken = read_stream(
                    ole_file, "ImageInfo/ImagesTaken", XDT.XRM_INT
                )[0]
                expected_number_frames = read_stream(
                    ole_file, "ImageInfo/NoOfImages", XDT.XRM_INT
                )[0]
                # Returns true even if all frames aren't written, throwing warning.
                if number_frames_taken != expected_number_frames:
                    logging.warning(
                        "%s is an incomplete %s file: only %i out of %i frames have been written",
                        path.name,
                        path.suffix,
                        number_frames_taken,
                        expected_number_frames,
                    )
                # Check for reference frame:
                if not ole_file.exists("ReferenceData/Image"):
                    logging.warning("No reference data found in file %s", path)
                return True
        else:
            logging.warning("Could not read ole file %s", path)
    else:
        logging.warning("%s not .txrm or .xrm", path)
    return False


def manual_save(
    filepath: Union[str, PathLike],
    image: Union[np.ndarray, List[np.ndarray]],
    data_type: Optional[DTypeLike] = None,
    metadata: Optional[OMEXML] = None,
):
    filepath = Path(filepath)
    image = np.asarray(image)
    tiff_kwargs = {}
    if data_type is not None:
        image = cast_to_dtype(image, data_type)
    else:
        logging.info("No data type specified. Saving with default data type.")
    if metadata is not None:
        meta_img = metadata.image()
        pixels = meta_img.Pixels
        pixels.set_PixelType(dtype_dict[image.dtype.name])
        meta_img.set_Name(filepath.name)
        resolution = [pixels.get_PhysicalSizeX(), pixels.get_PhysicalSizeY()]
        if None not in resolution:  # Check that both have values
            resolution = [1.e7 / (pixel_size)  for pixel_size in resolution]  # Pixels per resolution unit (converted from nm to cm)
            resolution_unit = tf.TIFF.RESUNIT.CENTIMETER  # Must use CENTIMETER for maximum compatibility
            if tf.__version__ >= "2022.7.28":
                # 2022.7.28: Deprecate third resolution argument on write (use resolutionunit)
                tiff_kwargs["resolutionunit"] = resolution_unit
            else:
                resolution.append(resolution_unit)
            tiff_kwargs["resolution"] = tuple(resolution)
        metadata = metadata.to_xml().encode()

    num_frames = len(image)
    bigtiff = (
        image.size * image.itemsize >= np.iinfo(np.uint32).max
    )  # Check if data bigger than 4GB TIFF limit

    logging.info("Saving image as %s with %i frames", filepath.name, num_frames)

    if filepath.exists():
        logging.warning("Overwriting existing file %s", filepath)

    with tf.TiffWriter(str(filepath), bigtiff=bigtiff, ome=False, imagej=False) as tif:
        tif.write(
            image,
            photometric="MINISBLACK",
            description=metadata,
            metadata={"axes": "ZYX"},
            software=f"txrm2tiff {__version__}",
            **tiff_kwargs
        )


def manual_annotation_save(
    filepath: Union[str, PathLike], image: Union[np.ndarray, List[np.ndarray]]
):
    filepath = Path(filepath)
    num_frames = len(image)
    logging.info(
        "Saving annotated image as %s with %i frames", filepath.name, num_frames
    )
    bigtiff = (
        image.size * image.itemsize >= np.iinfo(np.uint32).max
    )  # Check if data bigger than 4GB TIFF limit

    with tf.TiffWriter(str(filepath), bigtiff=bigtiff, ome=False) as tif:
        tif.write(
            image,
            photometric="RGB",
            metadata={"axes": "ZYXC"},
            software=f"txrm2tiff {__version__}",
        )



from random import randint
import unittest
from unittest.mock import MagicMock, patch
from numpy.testing import (
    assert_array_equal,
    assert_array_almost_equal,
    assert_approx_equal,
)

import numpy as np

from txrm2tiff.utils.image_processing import (
    cast_to_dtype,
    dynamic_despeckle_and_average_series,
    stitch_images,
    rescale_image,
)


class TestImageProcessing(unittest.TestCase):
    def test_stitch_mosaic(self):
        mosaic_xy_shape = (3, 4)
        image_size = (400, 400)
        fast_axis = 1
        slow_axis = 1 - fast_axis
        images = np.zeros(
            (
                mosaic_xy_shape[0] * mosaic_xy_shape[1],
                image_size[1],
                image_size[0],
            )
        )
        range_array = np.repeat(
            np.arange(0, mosaic_xy_shape[1]), image_size[0] * image_size[1]
        )
        images.flat = list(range_array) * mosaic_xy_shape[0]

        output_image = stitch_images(images, mosaic_xy_shape)

        expected_array = np.concatenate(
            [
                np.concatenate(
                    [np.full(image_size, i) for i in range(0, mosaic_xy_shape[1])],
                    axis=1,
                )
                for _ in range(0, mosaic_xy_shape[0])
            ],
            axis=0,
        )[np.newaxis, :, :]

        assert_array_equal(output_image, expected_array)

    def test_despeckle_ave(self):
        custom_reference = []
        dims = (250, 250)
        speckle_per_frame = 5
        mid_point = 10
        for i in range(0, mid_point * 2 + 1):
            speckle_array = np.full(dims, i)
            for _ in range(speckle_per_frame):
                speckle_idx = (randint(0, dims[0] - 1), randint(0, dims[1] - 1))
                speckle_array[speckle_idx] = i * randint(
                    500, 1000
                )  # Should be well beyond 2.8 standard devs
            custom_reference.append(speckle_array)
        custom_reference = np.asarray(custom_reference)
        expected_approx_output = np.full(dims, mid_point, dtype=np.float32)
        with self.assertRaises(AssertionError):
            assert_array_almost_equal(
                custom_reference, expected_approx_output, decimal=0
            )  # Should not be almost equal until despeckle & averaging
        assert_array_almost_equal(
            dynamic_despeckle_and_average_series(custom_reference),
            expected_approx_output,
            decimal=0,
        )

    def test_rescale_image(self):
        original_minimum = -6000
        original_maximum = 6000
        step_size = 500

        target_minimum = -600
        target_maximum = 600
        expected_step_size = 50

        array = np.arange(original_minimum, original_maximum + step_size, step_size)
        output_array = rescale_image(
            array, target_minimum, target_maximum, original_minimum, original_maximum
        )
        self.assertEqual(output_array[0], target_minimum)
        self.assertEqual(output_array[-1], target_maximum)
        self.assertEqual(output_array[1] - output_array[0], expected_step_size)

    def test_rescale_image_keep_range(self):
        target_minimum = 0
        target_maximum = 255
        array = np.zeros((5, 5), np.uint8)
        array[0, 0] = 255

        output_array = rescale_image(array, target_minimum, target_maximum)
        self.assertTrue(np.all(output_array[1:] == target_minimum))
        self.assertEqual(output_array[0, 0], target_maximum)

    def test_cast_to_dtype_with_no_change(self):
        dtype = np.float64
        info = np.finfo(dtype)
        min, max = info.min, info.max
        image = np.asarray([min, max], dtype=dtype)
        output = cast_to_dtype(image, dtype)
        assert_array_equal(image, output)

    def test_cast_to_dtype_with_float64_to_uint16(self):
        dtype = np.float64
        info = np.finfo(dtype)
        min, max = info.min, info.max
        image = np.asarray([min, max], dtype=dtype)

        target_dtype = np.uint16
        target_info = np.iinfo(target_dtype)
        target_min, target_max = target_info.min, target_info.max
        output = cast_to_dtype(image, target_dtype)
        self.assertEqual(output.dtype, target_dtype)
        self.assertEqual(output.min(), target_min)
        self.assertEqual(output.max(), target_max)

    def test_cast_to_dtype_with_float64_to_float32(self):
        dtype = np.float64
        info = np.finfo(dtype)
        min, max = info.min, info.max
        image = np.asarray([min, max], dtype=dtype)
        target_dtype = np.float32
        target_info = np.finfo(target_dtype)
        target_min, target_max = target_info.min, target_info.max
        output = cast_to_dtype(image, target_dtype)
        self.assertEqual(output.dtype, target_dtype)
        self.assertEqual(output.min(), target_min)
        self.assertEqual(output.max(), target_max)

    def test_cast_to_dtype_with_int32_to_int16(self):
        dtype = np.int32
        info = np.iinfo(dtype)
        min, max = info.min, info.max
        count = 50
        image = np.linspace(min, max, count, endpoint=True, dtype=dtype)

        target_dtype = np.int16
        target_info = np.iinfo(target_dtype)
        target_min, target_max = target_info.min, target_info.max
        output = cast_to_dtype(image, target_dtype)
        self.assertEqual(output.dtype, target_dtype)
        self.assertEqual(output.min(), target_min)
        self.assertEqual(output.max(), target_max)

    def test_cast_to_dtype_with_int32_to_float32(self):
        dtype = np.int32
        info = np.iinfo(dtype)
        min, max = info.min, info.max
        count = 50
        image = np.linspace(min, max, count, endpoint=True, dtype=dtype)

        target_dtype = np.float32
        target_info = np.finfo(target_dtype)
        sig_figs = target_info.iexp

        output = cast_to_dtype(image, target_dtype)
        self.assertEqual(output.dtype, target_dtype)

        # Should only affect min and max significant figures
        assert_approx_equal(
            output.min(),
            min,
            significant=sig_figs,
            err_msg=f"Min is not equal to {sig_figs} sig figs",
        )
        assert_approx_equal(
            output.max(),
            max,
            significant=sig_figs,
            err_msg=f"Max is not equal to {sig_figs} sig figs",
        )

    def test_cast_to_dtype_with_int32_to_float64(self):
        dtype = np.int32
        info = np.iinfo(dtype)
        min, max = info.min, info.max
        count = 50
        image = np.linspace(min, max, count, endpoint=True, dtype=dtype)

        target_dtype = np.float64

        output = cast_to_dtype(image, target_dtype)
        self.assertEqual(output.dtype, target_dtype)
        self.assertEqual(output.min(), min)  # Should not affect min and max
        self.assertEqual(output.max(), max)



import logging
from typing import List
import numpy as np

from .txrm_property import txrm_property


class ShiftsMixin:
    @txrm_property(fallback=None)
    def has_shifts(self) -> bool:
        return (
            self.has_stream("Alignment/X-Shifts")
            and self.has_stream("Alignment/Y-Shifts")
            and self.shifts_applied
        )

    @txrm_property(fallback=None)
    def shifts_applied(self):
        return np.any(self.x_shifts) or np.any(self.y_shifts)

    @txrm_property(fallback=None)
    def x_shifts(self) -> List:
        return self.read_stream("Alignment/X-Shifts")

    @txrm_property(fallback=None)
    def y_shifts(self) -> List:
        return self.read_stream("Alignment/Y-Shifts")

    def apply_shifts_to_images(self, images: np.ndarray) -> np.ndarray:
        if not self.shifts_applied:
            # if all shifts are 0, return the original image
            return images
        logging.info("Applying shifts to images")
        num_images = len(images)
        # Trim any shifts for empty images
        x_shifts = self.x_shifts[:num_images]
        y_shifts = self.y_shifts[:num_images]

        output = np.zeros_like(images)
        for im, out, x_shift, y_shift in zip(images, output, x_shifts, y_shifts):
            # Round to nearest pixel and scale to image shape
            x_shift = int(round(x_shift)) % im.shape[1]
            y_shift = int(round(y_shift)) % im.shape[0]
            if x_shift == 0 and y_shift == 0:
                out[:] = im[:]
            else:
                out[:] = np.roll(im, (y_shift, x_shift), axis=(0, 1))

        return output
